from pydoc import doc
import ipywidgets as widgets # Loads the Widget framework.
from IPython.core.magics.namespace import NamespaceMagics # Used to query namespace.
import os
import shutil
import re
import psutil
import subprocess as sp
import os
from threading import Thread , Timer
import sched, time
import json
from ipysheet import from_dataframe
import ipysheet
from timeit import default_timer as timer



class Monitor(Thread):
    def __init__(self,delay,path):
        super(Monitor,self).__init__()
        self.stopped=False
        self.delay=delay
        self.path=path
        self.memory={'gpu':[],'ram':[]}
        self.base_ram=self.get_ram_usage()
        self.base_gpu=self.get_gpu_memory()
        self.start_time=timer()
        self.start()
    
    def run(self):
        
        
        while not self.stopped:
            gpu=self.get_gpu_memory()
            ram=self.get_ram_usage()
            abs_gpu=0 if gpu-self.base_gpu<0 else gpu-self.base_gpu
            abs_ram=0 if ram-self.base_ram<0 else ram-self.base_ram
            self.memory['gpu'].append(abs_gpu)
            self.memory['ram'].append(abs_ram)
            time.sleep(self.delay)
            
    def get_ram_usage(self):
        
        """
        Obtains the absolute number of RAM bytes currently in use by the system.
        :returns: System RAM usage in bytes.
        :rtype: int

        """
        return int(int(psutil.virtual_memory().total - psutil.virtual_memory().available) / 1024 / 1024)


    def get_gpu_memory(self):
        
        output_to_list = lambda x: x.decode('ascii').split('\n')[:-1]
        ACCEPTABLE_AVAILABLE_MEMORY = 1024
        COMMAND = "nvidia-smi --query-gpu=memory.used --format=csv"
        try:
            memory_use_info = output_to_list(sp.check_output(COMMAND.split(),stderr=sp.STDOUT))[1:]
        except sp.CalledProcessError as e:
            raise RuntimeError("command '{}' return with error (code {}): {}".format(e.cmd, e.returncode, e.output))
        memory_use_values = [int(x.split()[0]) for i, x in enumerate(memory_use_info)]

        return memory_use_values[0]
    
    def stop(self):
        total_time=timer()-self.start_time
        self.stopped=True
        total_ram_consumption=max(self.memory['ram'])-min(self.memory['ram'])
        total_gpu_consumption=max(self.memory['gpu'])-min(self.memory['gpu'])
        self.memory['max_gpu_consumption']=total_gpu_consumption
        self.memory['max_ram_consumption']=total_ram_consumption
        self.memory['execution_time']=total_time
        if not os.path.exists(f"{self.path}/memory_info/"):
            os.makedirs(f"{self.path}/memory_info/")
        a_file = open(f"{self.path}/memory_info/memory_metrics.json", "w")
        a_file = json.dump(self.memory, a_file)
        
        
def copy(src,dst):
    path="/".join(dst.split("/")[:-1])
    if re.search('[a-zA-Z]',path):
        if not os.path.exists(path):
            os.makedirs(path)
    shutil.copy(src,dst)

def save_dict(dict_,file_type,file_path):
    #path="/".join(file_path.split("/")[:-1])
    #file_name=file_path.split("/")[-1]
    file_path=file_path.replace("\\","/")
    file_type=file_type.lower()
    types=["performance","prediction"]
    if file_type in types:
        file_path_type=file_path+"/"+file_type
        if not os.path.exists(file_path_type):
            os.makedirs(file_path_type)
        a_file = open(file_path_type+f"/{file_type}.json", "w")
        a_file = json.dump(dict_, a_file)
    elif(file_type=="none"):
        a_file = open(file_path, "w")
        a_file = json.dump(dict_, a_file)
    else:
        print("file type not one of [performance,prediction,none]")

def add2doc(EXP_PATH,input_):
    input_type=type(input_).__module__+" "+type(input_).__name__
    if input_type=='ipysheet.sheet Sheet':
        df=ipysheet.to_dataframe(input_)
        html=df.to_html()
        html=html.replace('<table border="1" class="dataframe">','<table border="1" width="100%" contenteditable=true')
        with open(f"{EXP_PATH}/document/doc_notes_docu.txt", 'a') as file:
            file.write(html+"\n")
    else:
        if not os.path.exists(f"{EXP_PATH}/document/"):
            os.mkdir(f"{EXP_PATH}/document/")  
            with open(f"{EXP_PATH}/document/doc_notes_docu.json", "w") as file1:
                json.dump({"contents":"","title":"Title"}, file1)
        elif not os.path.exists(f"{EXP_PATH}/document/doc_notes_docu.json"):
            with open(f"{EXP_PATH}/document/doc_notes_docu.json", "w") as file1:
                json.dump({"contents":"","title":"Title"}, file1)
        
        with open(f"{EXP_PATH}/document/doc_notes_docu.json", 'r') as file:
            doc_=json.load(file)
            contents=f"{doc_['contents']}<br>{input_}"
            doc_["contents"]=contents
        with open(f"{EXP_PATH}/document/doc_notes_docu.json", 'w') as file:
            json.dump(doc_, file)

def create_experiment(EXP_PATH):
    if not os.path.exists(EXP_PATH):
        os.makedirs(f"{EXP_PATH}/plots")


#!/usr/bin/env python
# coding: utf-8

# In[1]:


import plotly
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import os
import json
import pandas as pd
import numpy as np
from sklearn.metrics import roc_auc_score,accuracy_score,f1_score,precision_score,recall_score
import glob


# In[2]:


def get_exp_folders(exp_path):
    experiments=glob.glob(exp_path+"/*")
    experiments=list(map(lambda x:x.replace("\\","/"),experiments))
    experiments_list=list(map(lambda x:x.split("/")[-1],experiments))
    experiments_list_=[]
    experiments=[]
    for e in experiments_list:
        if not("." in e):
            experiments.append(f"{exp_path}/{e}")
            experiments_list_.append(e)
    return experiments,experiments_list_


# In[64]:


def get_metrics(exp_path):
    metrics_exp={}
    experiments,experiments_list_=get_exp_folders(exp_path)
    for i,exp in enumerate(experiments):
        prediction_path=f"{exp}/prediction/prediction.json"
        if os.path.exists(prediction_path):
            metrics_exp[experiments_list_[i]]={}
            pred_file = open(prediction_path, "r")
            pred_dict=json.load(pred_file)
            pred_file.close()
            y_true=np.array(pred_dict['y_true'])
            y_pred_proba=np.array(pred_dict['y_pred'])
            if(np.ndim(y_true)>1):
                y_true=np.argmax(y_true,axis=1)
                y_pred=np.argmax(y_pred_proba,axis=1)
                accuracy=accuracy_score(y_true,y_pred)*100
                f1score=f1_score(y_true,y_pred)*100
                precision=precision_score(y_true,y_pred)*100
                recall=recall_score(y_true,y_pred)*100
                metrics=['accuracy','f1score','precision','recall']
                for metric in metrics:
                    metrics_exp[experiments_list_[i]][metric]=round(eval(metric),3)
            else:
                y_pred=np.round(y_pred_proba)
                accuracy=accuracy_score(y_true,y_pred)
                f1score=f1_score(y_true,y_pred)
                precision=precision_score(y_true,y_pred)
                recall=recall_score(y_true,y_pred)
                metrics=['accuracy','f1score','precision','recall']
                for metric in metrics:
                    metrics_exp[experiments_list_[i]]['accuracy']=round(eval(metric),3)
    return metrics_exp    


# In[65]:


def get_memory(exp_path):
    memory_exp={}
    experiments,experiments_list_=get_exp_folders(exp_path)
    for i,exp in enumerate(experiments):
        memory_path=f"{exp}/memory_info/memory_metrics.json"
        if os.path.exists(memory_path):
            memory_exp[experiments_list_[i]]={}
            memory_file = open(memory_path, "r")
            memory_dict=json.load(memory_file)
            memory_exp[experiments_list_[i]]['ram']=memory_dict["max_ram_consumption"]
            memory_exp[experiments_list_[i]]['gpu']=memory_dict["max_gpu_consumption"]
            memory_exp[experiments_list_[i]]['execution_time']=memory_dict["execution_time"]
            memory_file.close()
    return memory_exp


# In[66]:


def plot_summary(exp_path):
    metrics_exp=get_metrics(exp_path)
    memory_exp=get_memory(exp_path)
    experiment_names_metrics = list(metrics_exp.keys())
    figs=[]
    if len(experiment_names_metrics)>0:
        fig1 = go.Figure()

        fig1.add_trace(go.Scatter(
            x=[metrics_exp[exp]['accuracy'] for exp in experiment_names_metrics],
            y=experiment_names_metrics,
            marker=dict(color="crimson", size=12),
            mode="markers",
            name="accuracy",
        ))

        fig1.add_trace(go.Scatter(
            x=[metrics_exp[exp]['f1score'] for exp in experiment_names_metrics],
            y=experiment_names_metrics,
            marker=dict(color="gold", size=12),
            mode="markers",
            name="f1 score"
        ))
        fig1.add_trace(go.Scatter(
            x=[metrics_exp[exp]['precision'] for exp in experiment_names_metrics],
            y=experiment_names_metrics,
            marker=dict(color="blue", size=12),
            mode="markers",
            name="precision"
        ))
        fig1.add_trace(go.Scatter(
            x=[metrics_exp[exp]['recall'] for exp in experiment_names_metrics],
            y=experiment_names_metrics,
            marker=dict(color="purple", size=12),
            mode="markers",
            name="recall"
        ))
        update_layout_(fig1,"Performance comparison","Percentage","Experiments")
        figs.append(fig1)
    experiment_names_memory = list(memory_exp.keys())
    if len(experiment_names_memory)>0:
        fig2 = go.Figure(data=[
            go.Bar(name='gpu', x=[memory_exp[exp]['gpu'] for exp in experiment_names_memory], y=experiment_names_memory,orientation='h'),
            go.Bar(name='ram', x=[memory_exp[exp]['ram'] for exp in experiment_names_memory], y=experiment_names_memory,orientation='h'),    
        ])
        fig3 = go.Figure(data=[
                go.Bar(name='execution time', x=[memory_exp[exp]['execution_time'] for exp in experiment_names_memory], y=experiment_names_memory,orientation='h')

        ])
        # Change the bar mode
        fig2.update_layout(barmode='group')
        update_layout_(fig2,"Memory Consumption","memory(MB)","Experiments")
        update_layout_(fig3,"Total execution time","time(seconds)","Experiments")
        figs.append(fig2)
        figs.append(fig3)
    return figs

def update_layout_(fig_,title,xaxis_title,yaxis_title):
    fig_.update_layout(
    title=title,
    xaxis_title=xaxis_title,
    yaxis_title=yaxis_title,
    font=dict(
        family="times new roman",
        size=16,
        color="black"
    )
)




import pandas as pd
from ipysheet import from_dataframe
import re
import ipysheet
import os
import pathlib

def open_sheet(source,rows=None,columns=None):
    type_=type(source).__name__
    
    if type_=='DataFrame':
        if rows:
            source=source.iloc[:rows,:]
        sheet=from_dataframe(source)
        sheet.column_width=[10]*len(source.columns)
        return sheet
    
    elif(".csv" in source):
        df=pd.read_csv(source)
        sheet=from_dataframe(df)
        return sheet
    
    elif(source=="new sheet"):
        rows=["" for _ in range(rows)]
        data={i:rows for i in list(map(chr, range(65, 65+columns)))}
        df=pd.DataFrame(data)
        sheet=ipysheet.from_dataframe(df)
        sheet.column_width=[10]*len(df.columns)
        return sheet
    
def download_sheet(sheet,file):
    
    path="/".join(file.split("/")[:-1])
    file_name=file.split("/")[-1].split(".")[0]
    experiment_name=file.split('/')[-2]
    if re.search('[a-zA-Z]',path):
        if not os.path.exists(path):
            os.makedirs(path)
    df=ipysheet.to_dataframe(sheet)
    if(file.endswith(".csv")):
        df.to_csv(file,index=False)
    elif(file.endswith(".xls")):
        html=df.to_html()
        html=html.replace('<table border="1" class="dataframe">','<table border="1" width="100%" contenteditable=true  class="table_" id="user_table" onclick="set_id_global(this.id)"')
        #if re.search('[a-zA-Z]',path):
        if not os.path.exists(path+f"/notes/"):
            os.makedirs(path+f"/notes/")
        with open(path+f"/notes/table_{experiment_name}.html", 'w') as f:
                f.write(html)
        with pd.ExcelWriter(file,mode='w') as writer:
                df.to_excel(writer,sheet_name="notes",index=False,header=True)
                
                
    



from flask import Flask,render_template,url_for,request,redirect
import plotly
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import os
import json
import pandas as pd
import numpy as np
from sklearn.metrics import roc_curve, roc_auc_score,precision_recall_curve,confusion_matrix
import jinja2
from pathlib import Path
import shutil
import pathlib
import glob
import re
from ml_track_tool.summary_plots import plot_summary
import imagesize
import seaborn as sns
import matplotlib.pyplot as plt
import base64
import io



def create_application(path):
    app=Flask(__name__,static_folder=f"{path}/plots")
    path=path.replace("\\","/")
    current_path=pathlib.Path(__file__).parent.resolve()
    if not os.path.exists(f"{path}/plots/logo"):
        os.makedirs(f"{path}/plots/logo")
    shutil.copy(f"{current_path}/logo.jpg",f"{path}/plots/logo")
    path="/".join(path.split("/")[:-1])
    current_path=pathlib.Path(__file__).parent.resolve()
    my_loader = jinja2.ChoiceLoader([
        app.jinja_loader,
        jinja2.FileSystemLoader(['/flaskapp/userdata',
                                 f"{current_path}/templates"]),
    ])
    app.jinja_loader = my_loader
    sns.set_theme()
    @app.route('/',methods=["POST","GET"])
    def home():   
        if request.method=='POST':
            if request.form['submit_btn']=="Show experiment":
                experiment_folder=request.form['experiment'] 
                return redirect(url_for("experiment_page",experiment=experiment_folder))
            elif request.form['submit_btn']=="Show Summary":
                figs=plot_summary(path)
                graphJSONs=[]
                if figs:
                    for i,fig in enumerate(figs):
                        graphJSON = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
                        graphJSONs.append((f"chart_{i}",graphJSON))
                    return render_template('summary.html', graphJSONs=graphJSONs)
                else:
                    return "<h1> No Info available</h1>"
        
        else:
            folders=os.listdir(path)
            return render_template("experiment_page.html",folders=folders)
        
    @app.route('/experiment_page/<experiment>',methods=["POST","GET"])
    def experiment_page(experiment):
        if request.method=='POST':
            download_files_txt=f"{Path.home()}/Downloads/*_docu.txt"
            download_files_json=f"{Path.home()}/Downloads/*_docu.json"
            doc_files_txt=glob.glob(download_files_txt)
            doc_files_json=glob.glob(download_files_json)
            doc_files_txt=list(map(lambda x:x.replace("\\","/"),doc_files_txt))
            doc_files_json=list(map(lambda x:x.replace("\\","/"),doc_files_json))
            if download_files_txt:
                for file in doc_files_txt:
                    file_name=file.split("/")[-1][:-9]
                    renamed_file=f"{Path.home()}/Downloads/{file_name}.txt"
                    os.rename(file,renamed_file)
                    shutil.move(renamed_file,f"{path+'/'+experiment}/plots/{file_name}.txt")
            if download_files_json:
                if not os.path.exists(f"{path+'/'+experiment}/document"):
                    os.makedirs(f"{path+'/'+experiment}/document")
                for file in doc_files_json:
                    file_name=file.split("/")[-1]
                    shutil.move(f"{Path.home()}/Downloads/{file_name}",f"{path+'/'+experiment}/document/{file_name}")

            notes=request.form.get("text_area_txt")
            text_file = open(f"{path}/{experiment}/team_notes/notes.txt", "w")
            text_file.write(notes)
            text_file.close()
            return redirect(url_for("experiment_page",experiment=experiment))
        else:
            exp_folder=experiment
            memory_path=f"{path+'/'+exp_folder}/memory_info/memory_metrics.json"
            history_path=f"{path+'/'+exp_folder}/performance/performance.json"
            prediction_path=f"{path+'/'+exp_folder}/prediction/prediction.json"
            plots_path=f"{path+'/'+exp_folder}/plots/"
            note_doc_path=f"{path+'/'+exp_folder}/document/*_docu.json"
            note_doc_files=glob.glob(note_doc_path)
            note_doc_files=list(map(lambda x:x.replace("\\","/"),note_doc_files))
            plot_files=glob.glob(f"{plots_path}/*.*")
            memory_file_path_exists=False
            history_file_path_exists=False
            prediction_file_path_exists=False
            plots_exists=False
            memory_dict={}
            history_dict={}
            pred_dict={}
            plot_lists=[]
            doc_contents=[]
            prev_num_doc=0
            docs_exists=False
            note_doc_exists=False
            plots_dict={}
            note_doc_dict={}
            if os.path.exists(memory_path):
                memory_file = open(memory_path, "r")
                memory_dict=json.load(memory_file)
                memory_file.close()
                fig, axes = plt.subplots(1, 2,figsize=(18,5))
                sns.lineplot(x=np.arange(len(memory_dict['gpu'])),y=memory_dict['gpu'],ax=axes[0])
                sns.lineplot(x=np.arange(len(memory_dict['ram'])),y=memory_dict['ram'],ax=axes[1])
                axes[0].set_title("GPU Consumption")
                axes[1].set_title("RAM Consumption")
                axes[0].set_ylabel("consumption (MB)")
                axes[0].set_xlabel("Time (s)")
                axes[1].set_ylabel("consumption (MB)")
                axes[1].set_xlabel("Time (s)")
                base64_plt=to_base64(plt)
                plots_dict['memory']=base64_plt.decode("utf-8")
                plots_dict['memory_exists']=True
            else:
                plots_dict['memory_exists']=False
            if os.path.exists(history_path):
                history_file = open(history_path, "r")
                history_dict=json.load(history_file)
                history_file.close()
                fig, axes = plt.subplots(1, 2,figsize=(18,5))
                sns.lineplot(x=np.arange(len(history_dict['val_categorical_accuracy'])),y=history_dict['val_categorical_accuracy'],ax=axes[0])
                sns.lineplot(x=np.arange(len(history_dict['categorical_accuracy'])),y=history_dict['categorical_accuracy'],ax=axes[0])
                sns.lineplot(x=np.arange(len(history_dict['loss'])),y=history_dict['loss'],ax=axes[1])
                sns.lineplot(x=np.arange(len(history_dict['val_loss'])),y=history_dict['val_loss'],ax=axes[1])
                axes[0].set_title("Accuracy")
                axes[0].set_ylabel("accuracy")
                axes[0].set_xlabel("Epochs")
                axes[1].set_ylabel("loss")
                axes[1].set_xlabel("epochs")
                base64_plt=to_base64(plt)
                plots_dict['history']=base64_plt.decode("utf-8")
                plots_dict['history_exists']=True
            else:
                plots_dict['history_exists']=False
            if os.path.exists(prediction_path):
                pred_file = open(prediction_path, "r")
                pred_dict=json.load(pred_file)
                pred_file.close()
                fig, axes = plt.subplots(1, 1,figsize=(9,5))
                y_true=np.argmax(pred_dict['y_true'],axis=1)
                y_pred=np.argmax(pred_dict['y_pred'],axis=1)
                cf=confusion_matrix(y_true,y_pred)
                sns.heatmap(cf,cbar=False,annot=True,ax=axes,fmt='g')
                axes.set_xlabel("Predicted")
                axes.set_ylabel("True labels")
                base64_plt=to_base64(plt)
                plots_dict['prediction']=base64_plt.decode("utf-8")
                plots_dict['prediction_exists']=True
            else:
                plots_dict['prediction_exists']=False
            if (note_doc_files):
                for file in note_doc_files:
                    if 'doc_notes_docu' not in file:
                        doc_id="_".join(file.split("/")[-1].split("_")[:2])
                        doc_file = open(file, "r")
                        doc_dict=json.load(doc_file)
                        doc_dict['id_']=doc_id
                        doc_contents.append(doc_dict)
                        docs_exists=True
                        doc_file.close()
                    else:
                        note_doc = open(file, "r")
                        note_doc_dict=json.load(note_doc)
                        note_doc_exists=True
                        note_doc.close()
                num_doc=list(map(lambda x:x.split("/")[-1][:-5].split("_")[1],note_doc_files))
                try:
                    prev_num_doc=int(max(x for x in num_doc if x.isdigit()))+1
                except Exception as e:
                    pass
            if not os.path.exists(f"{path}/{exp_folder}/team_notes/"):
                os.makedirs(f"{path}/{exp_folder}/team_notes/")
                text_file = open(f"{path}/{exp_folder}/team_notes/notes.txt", "w")
                text_file.write("")
                text_file.close()
            if plot_files:
                plots_exists=True
                plot_files=sorted(plot_files,key=os.path.getctime)
                plot_files=list(map(lambda x:x.replace("\\","/"),plot_files))
                plots_file_name=list(map(lambda x:x.split("/")[-1],plot_files))
                plot_lists=[]
                width,height=800,550
                for file in plots_file_name:
                    file_id=file.split(".")[0]
                    if (f"{file_id}.txt" in plots_file_name):
                        with open(f"{plots_path}/{file_id}.txt", "r+") as file1:
                            notes_str=file1.read()
                            notes_str_list=notes_str.split("{title}:")
                            notes_str=notes_str_list[0]
                            title=notes_str_list[-1] if len(notes_str_list)>1 else "Title" 
                            
                    else:
                            notes_str=""
                            title="Title"
                    if not file.endswith(".txt"):
                        with open(f"{path+'/'+exp_folder}/plots/{file}", "rb") as image_file:
                                encoded_img = base64.b64encode(image_file.read()).decode("utf-8")
                        width, height = imagesize.get(f"{path+'/'+exp_folder}/plots/{file}")
                    else:
                        continue
                    plot_lists.append((encoded_img,notes_str,file_id,width,f"save_note_{file_id}_txt",title))
            notes_path=f"{path}/{exp_folder}/team_notes/notes.txt"
            with open(notes_path) as f:
                team_notes = f.read()
            return render_template('visualization.html',img_list=plot_lists,plots_exist=plots_exists,page_title=experiment,doc_contents=doc_contents,note_doc=note_doc_dict,prev_num_doc=prev_num_doc,plots_dict=plots_dict,docs_exists=docs_exists,note_doc_exists=note_doc_exists,team_notes=team_notes)
    
    
    def to_base64(plt):
        my_stringIObytes = io.BytesIO()
        plt.savefig(my_stringIObytes, format='jpg')
        my_stringIObytes.seek(0)
        my_base64_jpgData = base64.b64encode(my_stringIObytes.read())
        return my_base64_jpgData
    
    app.run()
    
    


from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = '0.0.19'
DESCRIPTION = 'Documetation and Ml tracking tool using jupyter extensions'
LONG_DESCRIPTION = 'A package that allows users to document their ML experiments.'

# Setting up
setup(
    name="MlTrackTool",
    version=VERSION,
    author="Anesh",
    url="https://github.com/anesh-ml/Document-and-ML-track-tool",
    author_email="analytics955@gmail.com",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    package_dir={'':'.'},
    include_package_data=True,
    include_dirs=True,
    package_data={'ml_track_tool':["templates/*.html","logo.jpg"]},
    py_modules=["plotly_","summary_plots","OpenSheet","utils","flask_app"],
    install_requires=['ipysheet', 'scikit-plot', 'lime','Shapely','plotly','flask','jinja2','psutil','xlwt','imagesize'],
    keywords=['python', 'Machine learning', 'MLops', 'ML tracking system'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)


import plotly
import plotly.express as px
import plotly.subplots as sp


class px_utils():
    def __init__(self,fig):
        self.fig=fig if type(fig)=='list' else fig
        
    def make_subplot(self,subplots_titles,rows,cols,axis_titles,plot_title,legend_names,showlegend=True,width=None,height=None):
        '''
        Function to create subplots in plotly express
        input: 
            self.fig (iterables)
            subplot_title (iterables)
            rows (num of rows in subplot)
            cols (num of columns in subplot)
            axis_titles (iterables)
            plot_title: Title of the subplot
            width,height: width and height of the subplot

        usage example:
            subplot_px([fig,fig2],("temp","poll count"),1,2,[["temp","time"]],"Subplot express",900,600)
        '''
        if type(self.fig)!='list':
            fig_dict={}
            row=1
            col=1
            for i,f in enumerate(self.fig):
                for trace in range(len(f["data"])):
                    f['data'][0]['name']=legend_names[i]
                    f['data'][0]['showlegend']=showlegend
                    fig_dict[f"fig_{i}"]=f["data"][trace]
            this_figure = sp.make_subplots(rows=rows, cols=cols,subplot_titles=subplots_titles) 
            for traces in fig_dict.values():
                if (rows*cols>len(self.fig)):
                    this_figure.append_trace(traces, row=row, col=col)
                    '''
                    Assigns row and column for each plot.
                    '''
                    if col==cols:
                        row+=1
                        col=1
                        continue
                    col+=1
                else:
                    raise ValueError("number of subplot grids created is less than total figures.Check the row and col")
            '''
            setting axis titles
            '''
            i=1
            for ax_title in axis_titles:
                if i==1:
                    this_figure.layout[f'xaxis']['title']['text']=ax_title[0]
                    this_figure.layout[f'yaxis']['title']['text']=ax_title[1]
                else:
                    this_figure.layout[f'xaxis{i}']['title']['text']=ax_title[0]
                    this_figure.layout[f'yaxis{i}']['title']['text']=ax_title[1]
                i+=1
            this_figure.layout['title']['text']=plot_title
            if((height is not None)&(width is not None)):
                this_figure.update_layout(autosize=False,width=width,height=height)
        else:
            raise ValueError("input is not list of figures.")
        return this_figure

    def set_axis_title(self,axis_title):
        if(type(axis_title)!='list'):
            self.fig.layout['xaxis']['title']['text']=axis_title[0]
            self.fig.layout['yaxis']['title']['text']=axis_title[1]
        else:
            raise ValueError("input is not list of axis titles")
        return self.fig

    def set_plot_title(self,title):
        self.fig.layout.legend.title.text=title
        return self.fig
    
    def set_legend_name(self,legend_name):
        self.fig['data'][0]['name']=legend_name
        return self.fig